/*
 * GREAT-ER
 *
 * MODCALC.C
 *
 * model section - calculation routines
 *
 * author =  Geert Boeije
 * state =   27.05.1998
*/

#include <math.h>
#include <stdlib.h>
#include <malloc.h>
#include "init.h"
#include "modcalc.h"


/*
 * ***************************
 *
 * SOLUTION OF A LINEAR SYSTEM 
 *
 * ***************************
*/

/* after: Press WH, Teukolsky SA, Vetterling WT, Flannery BP (1992)
          Numerical Recipes in C - The Art of Scientific Computing - Second Edition
          Cambridge University Press
          ISBN 0-521-43108-5

          p. 43
*/

/*
 * ---------------------------------------------------------------------
 * Lower / Upper Diagonal Matrix decomposition
 * ---------------------------------------------------------------------
*/

int LU_Decomposition ( FloatPtr a_arg, 
                       int n_arg,
                       int* index_arg,
                       FloatPtr d_arg )
{
  int i, imax=0, j, k;
  FloatType big, dum, sum, temp;

  FloatPtr vv;    /* vv stores the implicit scaling of each row */

  vv = (FloatPtr) malloc ( n_arg * sizeof (FloatType) );

  /* set scaling vector to zero */
 
  for ( i=0 ; i < n_arg ; vv[i++]=0 );

  /* d_arg is used to indicate whether any rows have been interchanged */
  /* no row interchanges yet */  

  *d_arg = 1.0;

  /* loop over rows to get the implicit scaling information */

  for ( i=0 ; i<n_arg ; i++ )
  {
    big = 0.0;

    for ( j=0 ; j<n_arg ; j++ )
      if ( (temp = fabs( a_arg[i*n_arg + j] )) > big )
        big = temp;

    /* no non-zero largest element */    
    /* i.e. a singular matrix */

    if ( big == 0.0 )
    {
      free (vv);
      return 0;
    }


    /* save the scaling */

    vv[i] = 1.0 / big;
  }

  /* loop over colums of Crout's method */

  for ( j=0 ; j<n_arg ; j++ )
  {
    for ( i=0 ; i<j ; i++ )
    {
      sum = a_arg[i*n_arg + j];

      for ( k=0 ; k<i ; k++ )
        sum -= a_arg[i*n_arg + k] * a_arg[k*n_arg + j];

      a_arg[i*n_arg + j] = sum;
    }

    big = 0.0;

    for ( i=j ; i<n_arg ; i++ )
    {
      sum = a_arg[i*n_arg + j];

      for ( k=0 ; k<j ; k++ )
        sum -= a_arg[i*n_arg + k] * a_arg[k*n_arg + j];

      a_arg[i*n_arg + j] = sum;

      /* is the figure of merit for the pivot 
         better than the best so far ? */

      if ( (dum = vv[i]*fabs(sum)) >= big )
      {
        big = dum;
        imax = i;
      }
    }

    /* do we need to interchange rows ? */

    if ( j != imax )
    {
      /* interchange rows ! */
      for ( k=0 ; k<n_arg ; k++ )
      {
        dum = a_arg[imax*n_arg + k];
        a_arg[imax*n_arg + k] = a_arg[j*n_arg + k];
        a_arg[j*n_arg + k] = dum;
      }

      /* change the parity of d */
      *d_arg = -(*d_arg);         

      /* also interchange the scale factor 
         note: we no longer need vv[j] furtheron */
      vv[imax] = vv[j];
    }

    index_arg[j] = imax;

    /* for singular matrices: make the pivot element non-zero */

    if ( a_arg[j*n_arg + j] == 0.0 )
      a_arg[j*n_arg + j] = TINY;

    /* finally, divide the column by the pivot element */

    if ( j != n_arg-1 )
    {
      dum = 1.0 / (a_arg[j*n_arg + j]);

      for ( i=j+1 ; i<n_arg ; i++ )
        a_arg[i*n_arg + j] *= dum;
    }
  }

  free (vv);

  return 1; /* success */
}


/*
 * ---------------------------------------------------------------------
 * Solving the system by means of forward substitution and
 * backsubstitution, making use of the LU decomposed matrix
 * ---------------------------------------------------------------------
*/

void LU_Solver ( FloatPtr a_arg,
                 int n_arg,
                 int* index_arg,
                 FloatPtr b_arg )
{
  int i, ip, j;
  int ii=-1;      /* this is different from Numerical Recipes: 
                     ii can be zero if element[0] is accessed ! */
  FloatType sum;

  /* forward substitution */

  for ( i=0 ; i<n_arg ; i++ )
  {
    ip = index_arg[i];
    sum = b_arg[ip];
    b_arg[ip] = b_arg[i];

    if ( ii != -1 )
      for ( j=ii ; j<=i-1 ; j++ )
        sum -= a_arg[i*n_arg + j] * b_arg[j];
    else
      if ( sum)
        ii = i;

    b_arg[i] = sum;
  }

  /* backward substitution */

  for ( i=n_arg-1 ; i>=0 ; i-- )
  {
    sum = b_arg[i];

    for ( j=i+1 ; j<n_arg ; j++ )
      sum -= a_arg[i*n_arg + j] * b_arg[j];

    b_arg[i] = sum / a_arg[i*n_arg + i];
  }
}

/*
 * ---------------------------------------------------------------------
 * Solution of linear system A * X = B
 *   arguments = A, B, X and n          (with A - n x n matrix)
 * ---------------------------------------------------------------------
*/

int SolveLinearSystem ( FloatPtr a_arg,
                        FloatPtr b_arg,
                        FloatPtr x_arg,
                        int n_arg )
{
  int i,j, max;
  int* index;
  FloatType d;
  FloatType sum;
  FloatPtr a1;
  FloatPtr b1;

  max = (n_arg+1)*(n_arg+1);

  index = (int*) malloc ( n_arg * sizeof(int) );
  a1 = (FloatPtr) malloc ( max * sizeof(FloatType) );
  b1 = (FloatPtr) malloc ( n_arg * sizeof(FloatType) );

  for ( i=0 ; i<max ; i++ )
    a1[i] = a_arg[i];

  for ( i=0 ; i<n_arg ; i++ )
    b1[i] = b_arg[i];

  if (!LU_Decomposition ( a1, n_arg, index, &d ))
  {
    free (a1);
    free (b1);
    free (index);
    return 0;
  }

  LU_Solver ( a1, n_arg, index, b1 );

  /* control ! */

  for ( i=0 ; i<n_arg ; i++ )
  {
    sum = 0;
    for ( j=0 ; j<n_arg ; j++ )
      sum += a_arg[i*n_arg + j]*b1[j];

    if ( fabs ( sum - b_arg[i] ) > ACCURACY )
    {
      free (a1);
      free (b1);
      free (index);
      return 0;
    }

    x_arg[i] = b1[i];
  }

  free (a1);
  free (b1);
  free (index);

  return 1;
}

