/*
 * GREAT-ER
 *
 * MODEL.C
 *
 * model section
 *
 * author =  Geert Boeije
 *
 * changes by Intevation, http://intevation.net
 * state =   16.02.2000
 *
 * Intevation: replaced IN and OUT by IN__ and OUT__
*/

#include "init.h"
#include "r_river.h"
#include "r_wwtp.h"
#include "r_misc.h"
#include "modutil.h"
#include "data_out.h"
#include "usriface.h"
#include "navigate.h"
#include "names.h"
#include "model.h"

/* globals defined and declared here */

FloatType RiverInternal[NR_STATEVARS];
FloatPtr  SegmentRivClass;
FloatType DryWeatherFlow;
FloatType WWTP_DryWeatherFlow;
FloatType WaterUse;
Map_Ptr   ThisSegmentMap;
FloatType ThisSegmentID;
ChemProp  CP;
FloatType FugacityAir;
FloatType FugacityWater;
FloatType FugacityML;
FloatType FugacityPrimSludge;
FloatType DOSaturationML;
EmissionParPtr Em;

/*
 ********************************************************
 *
 * calculation of non-georeferenced chemical properties
 *
 * only ONCE per SHOT !!!
 *
 ********************************************************
*/

void ChemicalProperties ( void )
{
  NonGeoRef_ParPtr ngr = NonGeoRefParams;
  Chemical_ParPtr chem = ChemicalParams;

  /* Kd in mixed liquor */

  if ( chem->Kd_ML != -1 )
    CP.Kd_ML = chem->Kd_ML;
  else
    CP.Kd_ML = ngr->f_oc_ML * chem->K_ow;

  /* Kd in sewage */

  if ( chem->Kd_sewage != -1 )
    CP.Kd_sewage = chem->Kd_sewage;
  else
    CP.Kd_sewage = ngr->f_oc_sewage * chem->K_ow;

  /* Henry's law constant and K_H */

  if ( chem->H != -1 )
    CP.H = chem->H;
  else
    /* note: solubility in mg/L is converted to mol/L */
    CP.H = chem->P_vapour / (chem->solubility / chem->MM);

  CP.K_H = CP.H / ( GAS_CONSTANT * ( ngr->temp_air + KELVIN ) );

  /* fixed fugacities */

  FugacityAir = 1.0 / ( GAS_CONSTANT * ( ngr->temp_air + KELVIN ) );
  FugacityWater = 1.0 / CP.H;
  FugacityML = CP.Kd_ML * ngr->ro_SS_ML / CP.H;
  FugacityPrimSludge = CP.Kd_sewage * ngr->ro_SS_sewage / CP.H;;

  /*saturation DO */

  DOSaturationML =   14.65 
                   - 4.10e-1 * ngr->temp_sewage
                   + 7.99e-3 * ngr->temp_sewage * ngr->temp_sewage
                   - 7.78e-5 * ngr->temp_sewage * ngr->temp_sewage * ngr->temp_sewage;

  DOSaturationML *= ( 1.0 - 9e-6 * ngr->ML_salinity );
}


/*
 **************************
 *
 * EMISSION MODELS
 *
 **************************
*/

void FindEmission ( void )
{
  /* this segment's emission data */

  if ( ThisSegmentMap->emission != -1 )
    Em = (EmissionParPtr)(&EmissionData[ ThisSegmentMap->emission * NR_EMISSION_PARAMS ]);
  else
    Em = 0;
}

/*
 * ---------------------------------------------------------------------
 * domestic emission model
 *
 * down-the-drain emission by consumers
 * ---------------------------------------------------------------------
*/

void Model_DomesticEmission ( FloatPtr process_par_arg,
                              FloatPtr general_par_arg )
{
  DomesticEm_ParPtr p = (DomesticEm_ParPtr) process_par_arg;
  FloatType temp[NR_STATEVARS];
  FloatType out[NR_SUB_DOM][NR_DEST_DOM];
  FloatType segment_market;
  int i, j;

  /* this segment's market data */

  if ( Em )
    segment_market = Em->consumption;
  else
    segment_market = 0;


  /* 

  parameters:
  
  inhabitants           cap         # inhabitants connected to discharge point
  dry_weather_flow      l/cap.d     domestic emission flow


  */

  WaterUse = p->dry_weather_flow  * 1.157407407407e-8;   /* remember this for the WWTP models */
                                                         /* m3/cap.s */

  /* calculations */

  DryWeatherFlow = p->inhabitants * p->dry_weather_flow / 86400000 ;

  temp[_Q_] = DryWeatherFlow;
  temp[_PHI_] = p->inhabitants * segment_market / 31536 ;

  /* fill in B matrix */

  for ( i=0 ; i<NrSub[DOM] ; ++i )
    for ( j=0 ; j<NR_STATEVARS ; ++j )
      B[ oX(j,DOM,EM,i) ] = temp[j];          /* * p->subflow[i][j];*/

  /* make out matrix - to be used in destination splitter */

  #if NR_DEST_DOM > 1
  MakeOutMoreMatrix ( (FloatPtr)out, (FloatPtr)p->out, NR_SUB_DOM, NR_DEST_DOM );
  #else
  MakeOutOneMatrix ( (FloatPtr)out, NR_SUB_DOM );
  #endif

  /* fill in A matrix */

  DestinationSplitter (DOM, (FloatPtr)out, 0);
}

/*
 * ---------------------------------------------------------------------
 * non-domestic emission model
 *
 * sources from industry or from agriculture
 * ---------------------------------------------------------------------
*/

void Model_NonDomesticEmission ( FloatPtr process_par_arg,
                                 FloatPtr general_par_arg )
{
  NonDomesticEm_ParPtr p = (NonDomesticEm_ParPtr) process_par_arg;
  FloatType out[NR_SUB_NONDOM][NR_DEST_NONDOM];
  FloatType segment_nondom;

  int i;

  /*

  parameters:
  
  flow[i]    m3/s    emission flow in submodel i

  */

  /* this segment's nondomestic emission data */

  if ( Em )
    segment_nondom = Em->nondom;
  else
    segment_nondom = 0;

  /* fill in B matrix */

  for ( i=0 ; i<NrSub[NONDOM] ; ++i )
  {
    B[ oX(_Q_,NONDOM,EM,i) ] = p->flow[i];
    B[ oX(_PHI_,NONDOM,EM,i) ] = segment_nondom;
  }

  /* make out matrix - to be used in destination splitter */

  #if NR_DEST_NONDOM > 1
  MakeOutMoreMatrix ( (FloatPtr)out, (FloatPtr)p->out, NR_SUB_NONDOM, NR_DEST_NONDOM );
  #else
  MakeOutOneMatrix ( (FloatPtr)out, NR_SUB_NONDOM );
  #endif

  /* fill in A matrix */

  DestinationSplitter (NONDOM, (FloatPtr)out, 0);
}

/*
 * ---------------------------------------------------------------------
 * runoff emission model
 * ---------------------------------------------------------------------
*/

void Model_RunoffEmission ( FloatPtr process_par_arg,
                            FloatPtr general_par_arg )
{
  RunoffEm_ParPtr p = (RunoffEm_ParPtr) process_par_arg;
  FloatType out[NR_SUB_RUNOFF][NR_DEST_RUNOFF];
  FloatType segment_runoff;
  int i;

  /*

  parameters:
  
  flow[i]    m3/s    emission flow

  */
  

  /* this segment's runoff emission data */

  if ( Em )
    segment_runoff = Em->runoff;
  else
    segment_runoff = 0;

  /* fill in B matrix */

  for ( i=0 ; i<NrSub[RUNOFF] ; ++i )
  {
    B[ oX(_Q_,RUNOFF,EM,i) ] = p->flow[i];
    B[ oX(_PHI_,RUNOFF,EM,i) ] = segment_runoff;
  }

  /* make out matrix - to be used in destination splitter */

  #if NR_DEST_RUNOFF > 1
  MakeOutMoreMatrix ( (FloatPtr)out, (FloatPtr)p->out, NR_SUB_RUNOFF, NR_DEST_RUNOFF );
  #else
  MakeOutOneMatrix ( (FloatPtr)out, NR_SUB_RUNOFF );
  #endif

  /* fill in A matrix */

  DestinationSplitter (RUNOFF, (FloatPtr)out, 0);
}


/*
 ****************************
 *
 * TRANSPORT / TREATMENT MODELS
 *
 ****************************
*/

/*
 * ------------------------------------------------------------------------------
 * generic transport / treatment model - used by all
 * ------------------------------------------------------------------------------
*/

void Model_GenericTransportTreatment ( int      process_arg,
                                       FloatPtr process_par_arg,
                                       FloatPtr general_par_arg,
                                       FloatPtr out_arg )
{
  FloatType remaining[NR_STATEVARS];
  FloatType removal, leak;
  int i, j;
  
  /* model calculations */

  for ( i=0 ; i<NrSub[process_arg] ; ++i )
  {
    /* if groundwater leaks occur BEFORE the process, calculate leaks now */
    
    if ( NonGeoRefParams->groundwater_conservative  &&  DestNrBypass[process_arg] >= 0 )
      leak = out_arg [i*NrDestinations[process_arg] + DestNrBypass[process_arg]];
    else
      leak = 0;

    /* go to the different submodels */

    switch (process_arg)
    {
      case ONSITE:
        removal = OnSiteRemoval ( &X[ oX(0,process_arg,IN__,i) ], 
                                  process_par_arg, 
                                  general_par_arg,
                                  leak, 
                                  i );
        break;

      case SEWER:
        removal = SewerRemoval ( &X[ oX(0,process_arg,IN__,i) ], 
                                 process_par_arg, 
                                 general_par_arg,
                                 leak, 
                                 i );
        break;

      case TREAT:
        removal = WWTPRemoval ( &X[ oX(0,process_arg,IN__,i) ],
                                process_par_arg,
                                general_par_arg,
                                leak, 
                                i );
        break;

      case SURFWAT:
        removal = SmallSurfaceWaterRemoval ( &X[ oX(0,process_arg,IN__,i) ], 
                                             process_par_arg, 
                                             general_par_arg,
                                             leak, 
                                             i );
        break;

      case GROUNDWAT:
        removal = GroundWaterRemoval ( &X[ oX(0,process_arg,IN__,i) ], 
                                       process_par_arg, 
                                       general_par_arg,
                                       leak, 
                                       i );
        break;
 
      case RIVER:
        removal = RiverRemoval ( &X[ oX(0,process_arg,IN__,i) ], 
                                 process_par_arg, 
                                 general_par_arg,
                                 leak, 
                                 i );
        break;

      default:
        removal = 0;
        break;
    }

    /* if groundwater leaks occur AFTER the process, calculate leaks */
    
    if ( !NonGeoRefParams->groundwater_conservative  &&  DestNrBypass[process_arg] >= 0 )
      leak = out_arg [i*NrDestinations[process_arg] + DestNrBypass[process_arg]];

    /* model calculations */

    remaining[_Q_] = (1 - leak);
    remaining[_PHI_] = (1 - removal) * (1 - leak);

    /* fill in A matrix - transformation from IN to OUT */

    for ( j=0 ; j<NR_STATEVARS ; ++j )
      A[ AsubMatrixSize*j 
         + Acoordinates ( oX(0,process_arg,OUT__,i) , oX(0,process_arg,IN__,i) ) ] 
      = remaining[j];
  }

  /* fill in A matrix - sending from OUT to Destination */

  if (NrDestinations[process_arg])
    DestinationSplitter ( process_arg, 
                          out_arg, 
                          (NonGeoRefParams->groundwater_conservative  
                           &&  DestNrBypass[process_arg] >= 0) );
}

/*
 * ---------------------------------------------------------------------
 * on-site treatment model
 * ---------------------------------------------------------------------
*/

void Model_OnSiteTreatment ( FloatPtr process_par_arg,
                             FloatPtr general_par_arg )
{
  /* OnSiteTreatment_ParPtr p = (OnSiteTreatment_ParPtr) process_par_arg; */
  FloatType out[NR_SUB_ONSITE][NR_DEST_ONSITE];

  /* make out matrix - to be used in destination splitter */

  #if NR_DEST_ONSITE > 1
  MakeOutMoreMatrix ( (FloatPtr)out, (FloatPtr)p->out, NR_SUB_ONSITE, NR_DEST_ONSITE );
  #else
  MakeOutOneMatrix ( (FloatPtr)out, NR_SUB_ONSITE );
  #endif

  Model_GenericTransportTreatment ( ONSITE, process_par_arg, 
                                    general_par_arg, (FloatPtr)out );
}

/*
 * ---------------------------------------------------------------------
 * Sewer model
 * ---------------------------------------------------------------------
*/

void Model_Sewer ( FloatPtr process_par_arg,
                   FloatPtr general_par_arg )
{
  Sewer_ParPtr p = (Sewer_ParPtr) process_par_arg;
  FloatType out[NR_SUB_SEWER][NR_DEST_SEWER];
  int i;

  /* make out matrix - to be used in destination splitter */

  #if NR_DEST_SEWER > 1
  MakeOutMoreMatrix ( (FloatPtr)out, (FloatPtr)p->out, NR_SUB_SEWER, NR_DEST_SEWER );
  #else
  MakeOutOneMatrix ( (FloatPtr)out, NR_SUB_SEWER );
  #endif

  /* calculate dry weather flow into WWTP,
     this variable will be used later on in the WWTP models ! */

  for ( WWTP_DryWeatherFlow = 0, i=0 ; i<NR_SUB_SEWER ; ++i )
    WWTP_DryWeatherFlow += X[ oX( _Q_, SEWER, IN__, i ) ] * out[i][TREATED];

  /* replace comb. sewer dry weather flow by corrected value for non-dry weather
     (the Q into the sewer was previously calculated as the sum of all
     water emissions into it, except rainfall 

     Note: here we assume that NO RUNOFF MODEL IS USED !!!         */

  /*                              actual combined sewer flow
     CSewer_Q_DWF is the ratio: -------------------------------
                                dry weather combined sewer flow   */


  X[ oX( _Q_, SEWER, IN__, COMBINED ) ] =   CSewer_Q_DWF[ ThisSegmentMap->rivclass ] 
                                        * X[ oX( _Q_, SEWER, IN__, COMBINED ) ];


  Model_GenericTransportTreatment ( SEWER, process_par_arg, 
                                    general_par_arg, (FloatPtr)out );
}

/*
 * ---------------------------------------------------------------------
 * Treatment model
 * ---------------------------------------------------------------------
*/

void Model_Treatment ( FloatPtr process_par_arg,
                       FloatPtr general_par_arg )
{
  /* Treatment_ParPtr p = (Treatment_ParPtr) process_par_arg; */
  FloatType out[NR_SUB_TREAT][NR_DEST_TREAT];

  /* make out matrix - to be used in destination splitter */

  #if NR_DEST_TREAT > 1
  MakeOutMoreMatrix ( (FloatPtr)out, (FloatPtr)p->out, NR_SUB_TREAT, NR_DEST_TREAT );
  #else
  MakeOutOneMatrix ( (FloatPtr)out, NR_SUB_TREAT );
  #endif

  Model_GenericTransportTreatment ( TREAT, process_par_arg, 
                                    general_par_arg, (FloatPtr)out );
}

/*
 * ---------------------------------------------------------------------
 * groundwater
 * ---------------------------------------------------------------------
*/

void Model_GroundWater ( FloatPtr process_par_arg,
                         FloatPtr general_par_arg )
{
  FloatType out[NR_SUB_GROUNDWAT][NR_DEST_GROUNDWAT];

  /* make out matrix - to be used in destination splitter */

  #if NR_DEST_GROUNDWAT > 1
  MakeOutMoreMatrix ( (FloatPtr)out, (FloatPtr)p->out, 
                      NR_SUB_GROUNDWAT, NR_DEST_GROUNDWAT );
  #else
  MakeOutOneMatrix ( (FloatPtr)out, NR_SUB_GROUNDWAT );
  #endif

  Model_GenericTransportTreatment ( GROUNDWAT, process_par_arg, 
                                    general_par_arg, (FloatPtr)out );
}

/*
 * ---------------------------------------------------------------------
 * surface water
 * ---------------------------------------------------------------------
*/

void Model_SurfaceWater ( FloatPtr process_par_arg,
                          FloatPtr general_par_arg )
{
  FloatType out[NR_SUB_SURFWAT][NR_DEST_SURFWAT];

  /* make out matrix - to be used in destination splitter */

  #if NR_DEST_SURFWAT > 1
  MakeOutMoreMatrix ( (FloatPtr)out, (FloatPtr)p->out, NR_SUB_SURFWAT, NR_DEST_SURFWAT );
  #else
  MakeOutOneMatrix ( (FloatPtr)out, NR_SUB_SURFWAT );
  #endif

  Model_GenericTransportTreatment ( SURFWAT, process_par_arg, 
                                    general_par_arg, (FloatPtr)out );
}

/*
 * ---------------------------------------------------------------------
 * river
 * ---------------------------------------------------------------------
*/

void Model_River ( FloatPtr process_par_arg,
                   FloatPtr general_par_arg )
{
  FloatType dummy_out[] = { 1 };                     /* no destinations at all */
  River_ParPtr p = (River_ParPtr) process_par_arg;

  /* first replace matrix calculation of Q_in by number from parameters set
     - we assume that the prediction of the hydrological model is better 
     than the simple water balance in this model !!!                       */

  X[ oX( _Q_, RIVER, IN__, DEFAULT ) ] = p->Q;

  /* now we can do the job */  

  Model_GenericTransportTreatment (RIVER, process_par_arg, general_par_arg, dummy_out );
}

/*
 **************************
 *
 * SEGMENT CONTROL
 *
 **************************
*/

void SegmentControlSection ( int       current_segment_arg,
                             FloatPtr  segment_par_arg,
                             FloatPtr  segment_rivclass_arg,
                             FloatPtr  segment_iface_arg,
                             FloatPtr  segment_summary_arg )
{
  General_ParPtr genpar = (General_ParPtr)segment_par_arg;  
  int i;
  int in_offset;
  int start_process;
  FloatType stretch_length;

  /* position the map pointer to the current segment's map */
  ThisSegmentMap = &(Map[current_segment_arg]);
  ThisSegmentID = genpar->segment_ID;

  if ( (int)genpar->discharge )
  {
    start_process = 0;
    FindEmission ();
  }
  else
    start_process = RIVER;                         

  /* store segment inflow information into B array */

  InflowIntoSegment ( segment_iface_arg, &segment_par_arg[ParametersArrayOffset[RIVER]]);

  /* sequentially call process models (fill in matrix A and B)
     and solve the equation X = AX + B                         */

  for ( i=start_process ; i<NR_PROCESSES ; ++i )
  {
    /* do matrix calculation X = AX + B for INPUT of current process */

    in_offset = oX(0,i,IN__,0);

    switch ( ModelClass[i] )
    {
      case TRANSTREAT:
        PerformSubmatrixOperation ( in_offset, NrSub[i], i, IN__ );
        break;

      case EMISSION:
        break;

      default: break;
    }

    /* call process models */

    switch (i)
    {
      case DOM:
        Model_DomesticEmission 
        ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );
        break;

      case NONDOM:
        Model_NonDomesticEmission 
        ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );
        break;

      case RUNOFF:
        Model_RunoffEmission 
        ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );
        break;

      case ONSITE:
        Model_OnSiteTreatment 
        ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );
        break;

      case SEWER:
        Model_Sewer 
        ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );
        break;

      case TREAT:
        Model_Treatment 
        ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );

        break;

      case GROUNDWAT:
        Model_GroundWater 
        ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );
        break;

      case SURFWAT:
        Model_SurfaceWater 
        ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );
        break;

      case RIVER:
        SegmentRivClass = segment_rivclass_arg;
        Model_River ( &segment_par_arg[ ParametersArrayOffset[i] ], segment_par_arg );
        break;

      default: break;
    }

    /* do matrix calculation X = AX + B for OUTPUT of current process */

    switch ( ModelClass[i] )
    {
      case TRANSTREAT:
        PerformSubmatrixOperation ( oX(0,i,OUT__,0), NrSub[i], i, OUT__ );
        break;

      case EMISSION:
        PerformSubmatrixOperation ( oX(0,i,EM,0), NrSub[i], i, EM );
        break;

      default: break;
    }
  }

  stretch_length = ((River_ParPtr)(&segment_par_arg[ ParametersArrayOffset[RIVER] ]))->length;

  CalculateInterfaceVars ( segment_iface_arg );

  CalculateSummaryVars ( segment_summary_arg, 
                         (int)(genpar->discharge),
                         stretch_length );

  LogSegmentResults ( genpar->segment_ID,
                      segment_iface_arg,
                      segment_summary_arg,
                      start_process );
}
