/*
 * GREAT-ER
 *
 * MTCARLO.C
 *
 * Monte Carlo Engine
 *
 * author =  Geert Boeije
 * state =   07.04.1999
*/

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include "init.h"
#include "navigate.h"
#include "data_out.h"
#include "random.h"
#include "mc-check.h"
#include "mtcarlo.h"

FloatType FlowPercentile;

/*
 * **********************************************************************
 * 
 * R A N D O M   N U M B E R    G E N E R A T O R 
 *
 * **********************************************************************
*/


/* from: Press WH, Teukolsky SA, Vetterling WT, Flannery BP (1992)
         Numerical Recipes in C - The Art of Scientific Computing - Second Edition
         Cambridge University Press
         ISBN 0-521-43108-5

         p. 282
*/


/* long period ( > 2*10^18 ) random number generator of L'Ecuyer with Bays-Durham 
   shuffle and added safeguards.  Returns a uniform random deviate between 0.0 and 
   1.0 (exclusive of the endpoint values).  Call with idum a negative integer to 
   initialize; thereafter, do not alter idum between successive deviates in a 
   sequence.  RNMX should approximate the largest floating point value that is 
   less than 1.  
*/

FloatType Ran2 ( long* idum_arg )
{
  int j;
  long k;
  static long idum2 = 123456789;
  static long iy = 0;
  static long iv[NTAB];
  FloatType temp;

  if ( *idum_arg <= 0 )
  {
    /* this is something G. Boeije added: */

    *idum_arg = (long)time(NULL);

    /* end of addidition */

    /* this was removed:

    if ( -(*idum_arg) < 1 )
      *idum_arg = 1;
    else
      *idum_arg = -(*idum_arg);

    */

    idum2 = *idum_arg;

    for ( j=NTAB + 7 ; j>=0  ; j-- )
    {
      k = (*idum_arg) / IQ1;
      *idum_arg = IA1 * (*idum_arg - k*IQ1) - k*IR1;

      if ( *idum_arg < 0 )
        *idum_arg += IM1;

      if ( j < NTAB )
        iv[j] = *idum_arg;
    }
    iy = iv[0];
  }

  k = (*idum_arg) / IQ1;
  *idum_arg = IA1 * (*idum_arg - k*IQ1) - k*IR1;

  if ( *idum_arg < 0 )
    *idum_arg += IM1;

  k = idum2 / IQ2;
  idum2 = IA2 * (idum2 - k*IQ2) - k*IR2;

  if ( idum2 < 0 )
    idum2 += IM2;

  j = iy / NDIV;
  iy = iv[j] - idum2;
  iv[j] = *idum_arg;

  if ( iy < 1 )
    iy += IMM1;

  if ( (temp = AM*iy) > RNMX )
    return RNMX;
  else
    return temp;
}

/*
 * -------------------------------------------------------------
 * get a random number between 0 and 1
 * uniform distribution
 * -------------------------------------------------------------
*/

FloatType RandomNumber ( void )
{
  static long idum[] = {-1};

  return ( Ran2 (idum) );
}

/*
 * **********************************************************************
 * 
 * D I S T R I B U T I O N     R O U T I N E S 
 *
 * **********************************************************************
*/

/*
 * -------------------------------------------------------------
 * convert normal parameters to 'sister' lognormal parameters 
 * -------------------------------------------------------------
*/

NormalParams LogNormalFromNormal ( FloatType mean_arg,
                                   FloatType sdev_arg )
{
  NormalParams lognormal;
  FloatType normal_vari;
  FloatType temp;

  normal_vari = sdev_arg * sdev_arg;

  lognormal.mean = exp ( mean_arg + 0.5 * normal_vari );

  temp = exp ( normal_vari ) - 1;
  if ( temp > 0 )
    lognormal.sdev = lognormal.mean * sqrt ( temp );
  else
    lognormal.sdev = 0;

  return ( lognormal );
} 

/*
 * -------------------------------------------------------------
 * convert lognormal parameters to 'sister' normal parameters
 * -------------------------------------------------------------
*/

NormalParams NormalFromLogNormal ( FloatType mean_arg,
                                   FloatType sdev_arg )
{
  NormalParams normal;
  FloatType normal_vari;

  if ( sdev_arg == 0 || mean_arg == 0 )
  {
    normal.sdev = 0;
    normal.mean = log (mean_arg);
  }
  else  
  {
    normal_vari = log ( ((sdev_arg * sdev_arg) / (mean_arg * mean_arg)) + 1 );
    normal.sdev = sqrt ( normal_vari );

    normal.mean = log ( mean_arg ) - 0.5 * normal_vari;
  }

  return ( normal );
}

/*
 * -------------------------------------------------------------
 * convert lognormal 5th percentile (= low 95th percentile)
 * to the standard deviation of this distribution 
 * -------------------------------------------------------------
*/

FloatType StDevFrom95thPercentile ( FloatType mean_arg,
                                    FloatType low95_arg )
{
  FloatType s, t1, t2, t3;
  FloatType a = 1.6449;

  if ( low95_arg == 0 || mean_arg == 0 )
    return 0;

  t1 = a*a - 2*log( low95_arg / mean_arg );
  t2 = sqrt( t1 ) - a;   
  t3 = exp ( t2*t2 ) - 1;

  s = mean_arg * sqrt ( t3 );

  return (s);
}

/*
 * -------------------------------------------------------------
 * get an n-th percentile value from a normal distribution
 * -------------------------------------------------------------
*/

FloatType PercentileFromNormal ( FloatType mean_arg,
                                 FloatType sdev_arg,
                                 FloatType perc_arg )
{
  FloatType p, alpha;
  FloatType t, tt, ttt;
  int flag;

  if ( sdev_arg == 0.0 ) return mean_arg;

  if ( perc_arg < 0 ) perc_arg = 0;
  if ( perc_arg > 1 ) perc_arg = 1;

  /* alpha is the n-th percentile value in the standard normal distribution N(0,1) */

  if (  perc_arg < 0.5 )
  {
    flag = 1;
    p = perc_arg;
  }
  else
  {
    flag = 0;
    p = 1 - perc_arg;
  }

  tt = -2 * log (p);
  t = sqrt (tt);
  ttt = tt * t;

  alpha = t - ( 2.515517 + 0.802853*t + 0.010328*tt ) 
              / ( 1 + 1.432788*t + 0.189269*tt + 0.001308*ttt );

  if ( flag )
    alpha = -alpha;

  return ( mean_arg + alpha * sdev_arg );
}


/*
 * -------------------------------------------------------------
 * get an n-th percentile value from a log-normal distribution
 * -------------------------------------------------------------
*/

FloatType PercentileFromLogNormal ( FloatType mean_arg,
                                    FloatType sdev_arg,
                                    FloatType perc_arg )
{
  NormalParams normal;

  if ( sdev_arg == 0.0 ) return mean_arg;

  normal = NormalFromLogNormal ( mean_arg, sdev_arg );
  
  return ( exp ( PercentileFromNormal ( normal.mean, normal.sdev, perc_arg ) ) );
}


/*
 * -------------------------------------------------------------
 * generate a random normal distribution
 * -------------------------------------------------------------
*/

/* from: Press WH, Teukolsky SA, Vetterling WT, Flannery BP (1992)
         Numerical Recipes in C - The Art of Scientific Computing - Second Edition
         Cambridge University Press
         ISBN 0-521-43108-5

         p. 289
*/

/* returns a normally distributed deviate with zero mean and unit variance: N(0,1)
   using the Box-Muller method
*/

FloatType GasDev ( void )
{
  static int iset = 0;
  static FloatType gset;
  FloatType fac, rsq, v1, v2;

  /* we don't have an extra deviate handy */

  if ( iset == 0 )
  {
    do
    {
      /* pick 2 uniform number in the square extending
         from -1 to +1 in each direction                */

      v1 = 2.0 * RandomNumber () - 1.0;
      v2 = 2.0 * RandomNumber () - 1.0;

      /* are they in the unit circle ? */

      rsq = v1*v1 + v2*v2;
    }
    while ( rsq >= 1.0 || rsq == 0.0 );   /* if not, try again */

    fac = sqrt ( -2.0 * log(rsq) / rsq );

    /* make Box-Muller transformation to het 2 normal deviates:
        v1*fac
        v2*fac
    */

    gset = v1 * fac;
    iset = 1;

    return ( v2 * fac );
  }

  /* we have an extra deviate handy */

  else
  {
    iset = 0;
    return ( gset );
  }
}


/*
 * **********************************************************************
 * 
 * M O N T E    C A R L O     R O U T I N E S 
 *
 * **********************************************************************
*/

/*
 * --------------------------------------
 * get a shot from a uniform distribution
 * --------------------------------------
*/

FloatType ShotFromUniform( FloatType min_arg,
                           FloatType max_arg )
{
  FloatType width, temp, rn;

  /* make sure min is truly minimum ! */

  if ( min_arg > max_arg )
  {
    temp = min_arg;
    min_arg = max_arg;
    max_arg = temp;
  }

  width = max_arg - min_arg;

  if ( NrMonteCarloShots == 1 ) 
    rn = 0.5;
  else
    rn = RandomNumber();

  return ( min_arg + width*rn );
}

/*
 * --------------------------------------
 * get a shot from a normal distribution
 * --------------------------------------
*/

FloatType ShotFromNormal ( FloatType mean_arg,
                           FloatType sdev_arg )
{
  if ( ! sdev_arg  || NrMonteCarloShots == 1 ) return mean_arg;

  /* GasDev is N(0,1) */

  return ( mean_arg + sdev_arg * GasDev() );
}

/*
 * --------------------------------------
 * get a shot from a log-normal distribution
 * --------------------------------------
*/

FloatType ShotFromLogNormal ( FloatType mean_arg,
                              FloatType sdev_arg )
{
  NormalParams normal;

  if ( ! sdev_arg  || NrMonteCarloShots == 1 ) return mean_arg;

  normal = NormalFromLogNormal ( mean_arg, sdev_arg ); 

  return ( exp ( ShotFromNormal ( normal.mean, normal.sdev ) ) );
}

/*
 * ----------------------------------------------------
 * get a shot from a flows (or flows type) distribution
 * ----------------------------------------------------
*/

FloatType ShotFromFlowType ( FloatType mean_arg,
                             FloatType sdev_arg )
{
  /* this is a special case !!!

     arguments: mean = mean flow
                sdev = std. deviation, calculated from Q95 low flow
                       (in the initializing steps)

     distribution = lognormal

     We don't take a random shot, but we calculate the nth percentile
     of the distribution, where n is fixed in FlowPercentile.
  */

  if ( ! sdev_arg ) return mean_arg;

  if ( NrMonteCarloShots != 1 )
    return ( PercentileFromLogNormal ( mean_arg, sdev_arg, FlowPercentile ));
  else
    return ( mean_arg );
}

/*
 * --------------------------------------
 * get a shot from a distribution
 * --------------------------------------
*/

FloatType ObtainShotFromDistribution ( FloatPtr distri_par_arg,
                                       int distri_type_arg )
{
  switch ( distri_type_arg )
  {
    case NO_DISTRIB:
      return *distri_par_arg;

    case NORMAL_DISTRIB:
      return ShotFromNormal ( *distri_par_arg, *(distri_par_arg + 1) );

    case LOGNORMAL_DISTRIB:
      return ShotFromLogNormal ( *distri_par_arg, *(distri_par_arg + 1) );

    case UNIFORM_DISTRIB:
      return ShotFromUniform ( *distri_par_arg, *(distri_par_arg + 1) );;

    case FLOWTYPE_DISTRIB:
      return ShotFromFlowType ( *distri_par_arg, *(distri_par_arg + 1) );

    default:
      return *distri_par_arg;
  }
}

/*
 * --------------------------------------
 * copy a shot from a distribution vector
 * --------------------------------------
*/

void MakeMonteCarloVector ( FloatPtr from_arg,
                            FloatPtr to_arg,
                            int nr_arg,
                            int* distri_type_arg,
                            int* distri_index_arg    )
{
  int i;

  for ( i=0 ; i<nr_arg ; ++i )
    to_arg[i] = ObtainShotFromDistribution ( &from_arg[ distri_index_arg[i] ],
                                             distri_type_arg[i] );
}

/*
 * --------------------------------------
 * Correlation-corrected sampling - from Flowtype distribution
 * --------------------------------------
*/


FloatType CorrShotFromNormal ( FloatType mean_arg,
                               FloatType sdev_arg,
                               FloatType corr_arg )
{
  FloatType std_norm_dev_1;
  FloatType std_norm_dev_2;
  FloatType corr_norm_dev;

  if ( NrMonteCarloShots == 1 ) return mean_arg;
  if ( ! sdev_arg ) return mean_arg;

  /* corr_arg is the correlation between FlowPercentile and this variable */

  std_norm_dev_1 = PercentileFromNormal ( 0, 1, FlowPercentile );
  std_norm_dev_2 = ShotFromNormal ( 0, 1 );

  corr_norm_dev =    std_norm_dev_1 * corr_arg  
                  +  std_norm_dev_2 * sqrt ( 1 - corr_arg*corr_arg );

  return ( mean_arg + corr_norm_dev * sdev_arg );
}


FloatType CorrShotFromLogNormal ( FloatType mean_arg,
                                  FloatType sdev_arg,
                                  FloatType corr_arg )
{
  NormalParams normal;

  if ( NrMonteCarloShots == 1 ) return mean_arg;
  if ( ! sdev_arg ) return mean_arg;

  normal = NormalFromLogNormal ( mean_arg, sdev_arg ); 

  return ( exp ( CorrShotFromNormal ( normal.mean, normal.sdev, corr_arg ) ) );
}

FloatType ObtainCorrelatedShot ( FloatPtr distri_par_arg,
                                 int distri_type_arg,
                                 FloatType corr_arg )
{
  switch ( distri_type_arg )
  {
    case NO_DISTRIB:
      return *distri_par_arg;

    case NORMAL_DISTRIB:
      return CorrShotFromNormal ( *distri_par_arg, *(distri_par_arg + 1), corr_arg );

    case LOGNORMAL_DISTRIB:
    case FLOWTYPE_DISTRIB:
      return CorrShotFromLogNormal ( *distri_par_arg, *(distri_par_arg + 1), corr_arg );

    /* not yet implemented: 
    case UNIFORM_DISTRIB:
    */

    default:
      return *distri_par_arg;
  }
}


/*
 * ---------------------------------------
 * Monte Carlo Engine
 * ---------------------------------------
*/

void GenerateMonteCarloShot ( FloatPtr allsegments_par_arg,
                              FloatPtr allsegments_rivclass_arg,
                              FloatType* flow_perc_arg )
{
  /*
   *
   * GENERAL PHILOSOPHY:
   *
   * 1. take a random shot of FlowPercentile (between 0 and 100%)
   * 
   * 2. go through distributed parameter sets
   *    and create random shots (ie. uncorrelated) -
   *    except for FlowType distributions, which are all 100%
   *    correlated to FlowPercentile.
   * 
   * 3. take correlation corrected shots for those params which 
   *    need this approach - note: correlation only to FlowPecentile !
   *
   */

  int segment, disch, i;
  int a_discharge;
  FloatPtr thissegment_par;
  FloatType stretch_ID;

  /* get a random flow percentile
     which will be used in ALL SEGMENTS !
     any percentile between 1 and 100 is chosen   */

  if ( NrMonteCarloShots == 1 )
    FlowPercentile = -1;         /* single shot = mean flow */
  else
    FlowPercentile = RandomNumber ();

  *flow_perc_arg = FlowPercentile;

  /* make a random shot for the non-georeferenced parameters */
  
  MakeMonteCarloVector ( Distri_NonGeoRefParams,
                         (FloatPtr)NonGeoRefParams,
                         NR_NON_GEOREF_PARAMS,
                         NonGeoRefDistriType,
                         NonGeoRefDistriParIndex  );

  /* correlation corrected re-sampling */

  /* needed for: - water temperature 
                 - air temperature 
                 - wind speed  */

  NonGeoRefParams->temp_water
  = ObtainCorrelatedShot( &(Distri_NonGeoRefParams 
                              [ NonGeoRefDistriParIndex [INDEX_TEMP_WATER] ]),
                          NonGeoRefDistriType [INDEX_TEMP_WATER],
                          NonGeoRefParams->temp_corr );

  NonGeoRefParams->temp_air
  = ObtainCorrelatedShot( &(Distri_NonGeoRefParams 
                              [ NonGeoRefDistriParIndex [INDEX_TEMP_AIR] ]),
                          NonGeoRefDistriType [INDEX_TEMP_AIR],
                          NonGeoRefParams->temp_corr );


  NonGeoRefParams->wind_speed
  = ObtainCorrelatedShot( &(Distri_NonGeoRefParams 
                              [ NonGeoRefDistriParIndex [INDEX_WINDSPEED] ]),
                          NonGeoRefDistriType [INDEX_WINDSPEED],
                          NonGeoRefParams->wind_corr );

  /* verify shot values */

  CheckNonGeoRefShot (); 


  /* make a random shot for the chemical parameters */
  
  MakeMonteCarloVector ( Distri_ChemicalParams,
                         (FloatPtr)ChemicalParams,
                         NR_CHEMICAL_PARAMS,
                         ChemicalDistriType,
                         ChemicalDistriParIndex  );

  /* verify shot values */

  CheckChemicalShot (); 

  /* go through all segments for the RIVER and DISCHARGE part */
  /* -------------------------------------------------------- */

  for ( segment=0 ; segment < NrSegments ; ++segment )
  {
    thissegment_par = &allsegments_par_arg[ SegmentParOffset [segment] ];

    if ( (int)((General_ParPtr)thissegment_par)->segment_type == THE_SEA )
      continue;

    /* make a shot for the river data set */
    /* ---------------------------------- */
    
    MakeMonteCarloVector ( &Distri_RiverParams[ segment * TotRiverDistriPar ],
                           thissegment_par,
                           NrParamsOnlyRiver,
                           RiverDistriType,
                           RiverDistriParIndex   );

    /* correlation corrected re-sampling */

    /* needed for: - v 
                   - depth
                   - SS
                   - DO 
                   - BOD */

    /* normally, v is of FLOWTYPE distrib, and hence the sampling was
       already 100% correlated with flow percentile;
       however, if another distribution was specified for v,
       we need to do the resampling */

    if ( RiverDistriType[INDEX_RIV_v] != FLOWTYPE_DISTRIB )
      thissegment_par[INDEX_RIV_v]
      = ObtainCorrelatedShot( &(Distri_RiverParams 
                                  [ segment * TotRiverDistriPar 
                                    + RiverDistriParIndex [INDEX_RIV_v] ]),
                              RiverDistriType [INDEX_RIV_v],
                              1.0 );   /* 100% corr */

    thissegment_par[INDEX_RIV_depth]
    = ObtainCorrelatedShot( &(Distri_RiverParams 
                                [ segment * TotRiverDistriPar 
                                  + RiverDistriParIndex [INDEX_RIV_depth] ]),
                            RiverDistriType [INDEX_RIV_depth],
                            1.0 );   /* 100% corr */

    thissegment_par[INDEX_RIV_SS]
    = ObtainCorrelatedShot( &(Distri_RiverParams 
                                [ segment * TotRiverDistriPar 
                                  + RiverDistriParIndex [INDEX_RIV_SS] ]),
                            RiverDistriType [INDEX_RIV_SS],
                            NonGeoRefParams->SS_corr );

    thissegment_par[INDEX_RIV_DO]
    = ObtainCorrelatedShot( &(Distri_RiverParams 
                                [ segment * TotRiverDistriPar 
                                  + RiverDistriParIndex [INDEX_RIV_DO] ]),
                            RiverDistriType [INDEX_RIV_DO],
                            NonGeoRefParams->DO_corr );

    thissegment_par[INDEX_RIV_BOD]
    = ObtainCorrelatedShot( &(Distri_RiverParams 
                              [ segment * TotRiverDistriPar 
                                + RiverDistriParIndex [INDEX_RIV_BOD] ]),
                            RiverDistriType [INDEX_RIV_BOD],
                            NonGeoRefParams->BOD_corr );

    /* verify shot values */

    CheckGeneralShot ( (General_ParPtr)thissegment_par );
    CheckRiverShot ( (River_ParPtr)(&thissegment_par[ ParametersArrayOffset[RIVER] ]) );

    /* if there is a discharge, make a shot of the discharge data set */
    /* -------------------------------------------------------------- */

    a_discharge = (int)(((General_ParPtr)thissegment_par)->discharge);

    if ( a_discharge )
    {
      stretch_ID = ((General_ParPtr)thissegment_par)->segment_ID;

      /* find the corresponding spot in the discharge data */
      
      disch = (&Map[ segment ])->discharge ;

      /* note: we have to skip the 1st value in discharge dataset because it is 
         the stretch ID - that's why we have the '+1'
      */
      
      MakeMonteCarloVector ( &Distri_DischargeParams[ disch*TotDischargeDistriPar + 1 ],
                             &thissegment_par[ NrParamsOnlyRiver ],
                             NrParamsDischarge - NrParamsOnlyRiver,
                             DischargeDistriType,
                             DischargeDistriParIndex );

      /* correlation corrected re-sampling */

      /* verify shot values */

      CheckDischargeShot ( thissegment_par );
    }
  }

  /* go through all EMISSION data points */
  /* --------------------------------- */

  for ( i=0 ; i < NrEmissions ; ++i )
  {
    thissegment_par = &EmissionData[ NR_EMISSION_PARAMS*i ];

    MakeMonteCarloVector ( &Distri_EmissionParams [ i * TotEmissionDistriPar ],
                           thissegment_par,
                           NR_EMISSION_PARAMS,
                           EmissionDistriType,
                           EmissionDistriParIndex  );

    /* correlation corrected re-sampling */

    /* verify shot values */

    CheckEmissionShot ( thissegment_par );
  }

  /* go through all WWTPs */
  /* -------------------- */

  for ( i=0 ; i < NrWWTPs ; ++i )
  {
    thissegment_par = &WWTPData[ NR_WWTP_PARAMS*i ];

    MakeMonteCarloVector ( &Distri_WWTPParams [ i * TotWWTPDistriPar ],
                           thissegment_par,
                           NR_WWTP_PARAMS,
                           WWTPDistriType,
                           WWTPDistriParIndex  );

    /* correlation corrected re-sampling */

    /* verify shot values */

    CheckWWTPShot ( thissegment_par );
  }

  /* go through all RivClasses */
  /* ------------------------- */

  for ( i=0 ; i < NrRivClasses ; ++i )
  {
    thissegment_par = &allsegments_rivclass_arg[ NR_RIVCLASS_PARAMS*i ];

    MakeMonteCarloVector ( &Distri_RivClassParams [ i * TotRivClassDistriPar ],
                           thissegment_par,
                           NR_RIVCLASS_PARAMS,
                           RivClassDistriType,
                           RivClassDistriParIndex  );

    /* correlation corrected re-sampling */

    /* needed for: - SS
                   - DO
                   - BOD
                   - k_SS                   */

    thissegment_par[INDEX_RC_SS]
    = ObtainCorrelatedShot( &(Distri_RivClassParams 
                                  [ i * TotRivClassDistriPar 
                                    + RivClassDistriParIndex [INDEX_RC_SS] ]),
                            RivClassDistriType [INDEX_RC_SS],
                            NonGeoRefParams->SS_corr );

    thissegment_par[INDEX_RC_DO]
    = ObtainCorrelatedShot( &(Distri_RivClassParams 
                                  [ i * TotRivClassDistriPar 
                                    + RivClassDistriParIndex [INDEX_RC_DO] ]),
                            RivClassDistriType [INDEX_RC_DO],
                            NonGeoRefParams->DO_corr );

    thissegment_par[INDEX_RC_BOD]
    = ObtainCorrelatedShot( &(Distri_RivClassParams 
                                  [ i * TotRivClassDistriPar 
                                    + RivClassDistriParIndex [INDEX_RC_BOD] ]),
                            RivClassDistriType [INDEX_RC_BOD],
                            NonGeoRefParams->BOD_corr );

    thissegment_par[INDEX_RC_k_SS]
    = ObtainCorrelatedShot( &(Distri_RivClassParams 
                                [ i * TotRivClassDistriPar 
                                  + RivClassDistriParIndex [INDEX_RC_k_SS] ]),
                            RivClassDistriType [INDEX_RC_k_SS],
                            NonGeoRefParams->k_SS_corr );

    /* verify shot values */

    CheckRivClassShot ( thissegment_par );

    /* set up the CSewer_Q_DWF array

       this contains - for each river class - the ratio between
       flow and dry weather flow in the combined sewer system

       this value depends on the correlation between Q_river and Q_wwtp,
       which is river class specific ! */

    CSewer_Q_DWF [i] 
    = ObtainCorrelatedShot( &(Distri_NonGeoRefParams 
                              [ NonGeoRefDistriParIndex [INDEX_CS_Q_DWF] ] ),
                            NonGeoRefDistriType [INDEX_CS_Q_DWF],
                            ((RivClass_ParPtr)thissegment_par)->corr_riv_csewer );

    /* verify CSewer_Q_DWF value */

    CheckCSewerQDWFShot (&(CSewer_Q_DWF[i]));
  }
}


/*
 * **********************************************************************
 * 
 * I N I T I A L I Z I N G    R O U T I N E S 
 *
 * **********************************************************************
*/

/*
 * --------------------------------------
 * modify a distribution vector
 * --------------------------------------
*/

void ModifyDistribVector ( FloatPtr vector_arg,
                           int nr_arg,
                           int* distri_type_arg,
                           int* distri_index_arg    )
{
  int i;
  FloatPtr this_one;

  for ( i=0 ; i<nr_arg ; ++i )
  {
    switch ( distri_type_arg[i] )
    {
      case FLOWTYPE_DISTRIB:
        this_one = &vector_arg[ distri_index_arg[i] ];
        if ( this_one[0] != 0.0 && this_one[1] != 0.0 )
          this_one[1] = StDevFrom95thPercentile ( this_one[0], this_one[1] );
        break;

      case NO_DISTRIB:
      case NORMAL_DISTRIB:
      case LOGNORMAL_DISTRIB:
      case UNIFORM_DISTRIB:
      default:
        break;
    }
  }
}

/*
 * ------------------------------------------------------------------------------
 * Initializing:
 *
 * produce st_dev values instead of Q95 low flows
 * in the FLOWTYPE_DISTRIB distributions (ie, flows and flow velocities) 
 * ------------------------------------------------------------------------------
*/

void InitializeFlowTypeDistrib ( void )
{
  int segment, disch;

  /* non-georeferenced parameters */
  
  ModifyDistribVector ( Distri_NonGeoRefParams,
                        NR_NON_GEOREF_PARAMS,
                        NonGeoRefDistriType,
                        NonGeoRefDistriParIndex  );

  /* go through all river segments */

  for ( segment=0 ; segment < NrSegments ; ++segment )
    ModifyDistribVector ( &Distri_RiverParams[ segment * TotRiverDistriPar ],
                          NrParamsOnlyRiver,
                          RiverDistriType,
                          RiverDistriParIndex   );

  /* go through all discharges */

  for ( disch=0 ; disch < NrSegmentsDischarge ; ++disch )
    ModifyDistribVector ( &Distri_DischargeParams[ disch * TotDischargeDistriPar + 1 ],
                          NrParamsDischarge - NrParamsOnlyRiver,
                          DischargeDistriType,
                          DischargeDistriParIndex   );

}



/*
 * **********************************************************************
 * 
 * R E S U L T S     P R O C E S S I N G
 *
 * **********************************************************************
*/


/* make sums of results and results^i */

void MakeSums ( FloatPtr allsegments_summary_arg,
                FloatPtr allsegments_sums_arg )
{
  int i, j;
  FloatPtr current;

  for ( i=0 ; i<AllSummarySize ; ++i )
  {
    current = &allsegments_sums_arg[ i*NR_SUMMARY_SUMS ];
    
    for ( j=0 ; j<NR_SUMMARY_SUMS ; ++j )
      current[ j ] 
        += pow ( allsegments_summary_arg[ i ], j+1 );
  }
}

/* calculate the results' distribution parameters */

void CalculateDistribution ( FloatPtr allsegments_sums_arg )
{
  int i, calc;
  int nr, seg;
  FloatPtr current_distri_seg, current_distri;
  FloatPtr current_sum_seg, current_sum;
  FloatType mean, variance;
  
  calc = NR_SUMMARY_SUMS * NR_SUMMARY_TERMS * NR_SUMMARY_VARS;

  for ( seg=0 ; seg<NrSegments ; ++seg )
  {
    current_distri_seg = &Distri_SummaryVars[ seg * TotSummaryDistriPar ]; 
    current_sum_seg = &allsegments_sums_arg[ seg * calc ];

    for ( i=0 ; i < NR_SUMMARY_VARS * NR_SUMMARY_TERMS ; ++i )
    {
      current_distri = &current_distri_seg[ SummaryDistriParIndex[ i ] ];
      current_sum = &current_sum_seg[ i * NR_SUMMARY_SUMS ];
      
      /* mean */
        
      mean = current_sum[ 0 ] / NrMonteCarloShots;

      switch ( SummaryDistriType[i] )
      {
        case NO_DISTRIB:

          current_distri[ 0 ] = mean;
          break;

        case NORMAL_DISTRIB:
        case LOGNORMAL_DISTRIB:
        case PECTYPE_DISTRIB:
        case FLOWTYPE_DISTRIB:

          nr = NrMonteCarloShots - 1;
          current_distri[ 0 ] = mean;

          if ( nr > 0 )
          {
            variance = ( current_sum[ 1 ] - NrMonteCarloShots * mean * mean ) / nr;
            current_distri[1] = ( variance > 0 ? sqrt ( variance ) : 0 );
          }
          else
            current_distri[1] = 0;

          break;

       default:
         break;
      }

      switch ( SummaryDistriType[i] )
      {
        case PECTYPE_DISTRIB:

          /* we want the 95th (high) PEC percentile instead of std. dev */ 
          current_distri[ 1 ] 
            = PercentileFromLogNormal ( current_distri[ 0 ], current_distri[ 1 ], 0.95 );
          break;

        case FLOWTYPE_DISTRIB:

          /* we want the 5th (low) flow percentile instead of std. dev */

          current_distri[ 1 ] 
            = PercentileFromLogNormal ( current_distri[ 0 ], current_distri[ 1 ], 0.05 );
          break;

        default:
          break;
      }
    }
  }
}


