/*
 * GREAT-ER
 *
 * NAVIGATE.C
 *
 * navigation through the X, A and B matrices and the parameters arrays
 *
 * author =  Geert Boeije
 *
 * changes by Intevation, http://intevation.net
 * state =   16.02.2000
 *
 * Intevation: replaced IN and OUT by IN__ and OUT__
*/

#include "init.h"
#include "navigate.h"

/* a structure which is used globally in this module */

static struct
{
  int modelclass;
  int terminal;
}                    RowType [ONE_STATEVAR_SIZE];

/*
 ********************************************************************
 * P R E P A R A T O R Y    S T E P S
 *
 * to be done just once
 ********************************************************************
*/

/*
 * ---------------------------------------------------------------------
 * calculate state variables array offset for each process
 * ---------------------------------------------------------------------
*/

void CalculateStateVarsArrayOffset (void)
{
  int i, plus;

  /* determine each process's offset within statevar list */

  StateVarsArrayOffset[0] = 0;

  for ( i=0 ; i < NR_PROCESSES ; ++i )
  {
    switch (ModelClass[i])
    {
      case EMISSION:

        plus = NrSub[i] * 1;
        break;

      case TRANSTREAT:

        plus = NrSub[i] * 2;
        break;

      default:

        plus = 0;
        break;
    }

    StateVarsArrayOffset[ i+1 ] = StateVarsArrayOffset[ i ] + plus;
  }
}

/*
 * ---------------------------------------------------------------------
 * calculation of A matrix offset and A size
 * ---------------------------------------------------------------------
*/

int GetNrTerminalsInProcess(int process_arg)
{
  int r;

  switch ( ModelClass[process_arg] )
  {
    case EMISSION:

      r = 1;
      break;

    case TRANSTREAT:

      r = 2;
      break;

    default:
  
      r = 0;
      break;
  }

  return r;
}

void CalculateAmatrixOffset (void)
{
  int i, process=0, proc_row=0, count, term_nr;

  /* A is a lower triangular matrix with zeros on diagonal
     Size of A excludes most zeros see report                 */

  term_nr = GetNrTerminalsInProcess(process);

  for ( i=0, AsubMatrixSize=0, count=1 ; i<OneStateVarSize ; i++ )
  {
    AsubMatrixOffset[i] = AsubMatrixSize;

    if ( ModelClass[process] == TRANSTREAT )
      if (count <= NrSub[process] )            /* input */
      {
        AsubMatrixSize += proc_row;
        RowType[i].terminal = IN__;
      }
      else                                     /* output */
      {
        AsubMatrixSize += 1;
        RowType[i].terminal = OUT__;
      }
    else
      RowType[i].terminal = 0;

    /* note: emission does everything via B array */

    RowType[i].modelclass = ModelClass[process];

    if ( ++count > NrSub[process] * term_nr )
    {
      count = 1;
      ++process;
      proc_row = i+1;
      term_nr = GetNrTerminalsInProcess(process);
    }
  }

  AmatrixSize = AsubMatrixSize * NR_STATEVARS;
}

/*
 * ---------------------------------------------------------------------
 * calculate the offset in the 'all segments' parameters array
 * for each segment 
 * ---------------------------------------------------------------------
*/

void CalculateSegmentParametersOffset ( void )
{
  General_ParPtr p = (General_ParPtr) Distri_RiverParams;
  int i, plus;
  
  SegmentParOffset[ 0 ] = 0;

  for ( i=1 ; i < NrSegments ; ++i )
  {
    if ( p->discharge )
      plus = NrParamsDischarge;
    else
      plus = NrParamsOnlyRiver;

    SegmentParOffset[ i ] = SegmentParOffset[ i-1 ] + plus;

    p = (General_ParPtr) (&Distri_RiverParams[ i * TotRiverDistriPar ]);
  }
}


/*
 * ---------------------------------------------------------------------
 * calculate the offset in the parameters array
 * ---------------------------------------------------------------------
*/

void CalculateParametersArrayOffset ( void )
{
  int i;

/*
   first the special case: RIVER
   RIVER is the last process, but its parameters come first
   (because it is the only process in some stretches)
*/

  ParametersArrayOffset[RIVER] = NR_GEOREF_GEN_PARAMS;

  ParametersArrayOffset[0] = ParametersArrayOffset[RIVER] + NrParamsInProcess[RIVER] + 1;
  /* the +1 is for the emission ID ! */

  for ( i=1 ; i < RIVER ; ++i )
    ParametersArrayOffset[i] = ParametersArrayOffset[i-1] + NrParamsInProcess[i-1];
}

/*
 * ---------------------------------------------------------------------
 * calculate the offset in the distribution parameters array
 * BEFORE MONTE CARLO
 * ---------------------------------------------------------------------
*/

void CalculateDistriParIndex (void)
{
  int i;

  /* non-geo-referenced parameters */

  NonGeoRefDistriParIndex[0] = 0;

  for  ( i=1 ; i < NR_NON_GEOREF_PARAMS ; ++i )
    NonGeoRefDistriParIndex[ i ] =   NonGeoRefDistriParIndex[ i-1 ] 
                                   + NrDistriParams[ NonGeoRefDistriType[ i-1 ] ];

  TotNonGeoRefDistriPar =   NonGeoRefDistriParIndex[ i-1 ]
                          + NrDistriParams[ NonGeoRefDistriType[ i-1 ] ];

  /* chemical parameters */

  ChemicalDistriParIndex[0] = 0;

  for  ( i=1 ; i < NR_CHEMICAL_PARAMS ; ++i )
    ChemicalDistriParIndex[ i ] =   ChemicalDistriParIndex[ i-1 ] 
                                   + NrDistriParams[ ChemicalDistriType[ i-1 ] ];

  TotChemicalDistriPar =   ChemicalDistriParIndex[ i-1 ]
                          + NrDistriParams[ ChemicalDistriType[ i-1 ] ];

  /* geo-referenced parameters - river */

  RiverDistriParIndex[0] = 0;

  for ( i=1 ; i < NrParamsOnlyRiver ; ++i )
    RiverDistriParIndex[i] =  RiverDistriParIndex[ i-1 ] 
                             + NrDistriParams[ RiverDistriType[ i-1 ] ];

  TotRiverDistriPar =   RiverDistriParIndex[ i-1 ] 
                      + NrDistriParams[ RiverDistriType [ i-1 ] ];

  /* geo-referenced parameters - discharge */

  DischargeDistriParIndex[0] = 0;

  for ( i=1 ; i < NrParamsDischarge - NrParamsOnlyRiver ; ++i )
    DischargeDistriParIndex[i] =  DischargeDistriParIndex[ i-1 ] 
                                + NrDistriParams[ DischargeDistriType[ i-1 ] ];

  TotDischargeDistriPar =   DischargeDistriParIndex[ i-1 ] 
                          + NrDistriParams[ DischargeDistriType [ i-1 ] ]
                          + 1; 
                          /* the +1 is for the stretch_ID
                             which come upfront each discharge */


  WWTP_ID_Location = DischargeDistriParIndex [WWTP_ID_Location] + BEFORE_WWTP_ID + 1;
  Emission_ID_Location = DischargeDistriParIndex [Emission_ID_Location] + 1;

  /* summary variables */
  
  SummaryDistriParIndex[0] = 0;
  
  for ( i=1 ; i < NR_SUMMARY_VARS * NR_SUMMARY_TERMS ; ++i )
    SummaryDistriParIndex[i] =   SummaryDistriParIndex[ i-1 ]
                               + NrDistriParams[ SummaryDistriType[ i-1 ] ];

  TotSummaryDistriPar =   SummaryDistriParIndex[ i-1 ]
                        + NrDistriParams[ SummaryDistriType[ i-1 ] ];

  /* emission data */

  EmissionDistriParIndex[0] = 0;

  for ( i=1 ; i < NR_EMISSION_PARAMS ; ++i )
    EmissionDistriParIndex[i] =   EmissionDistriParIndex[ i-1 ]
                                + NrDistriParams[ EmissionDistriType[ i-1 ] ];

  TotEmissionDistriPar =   EmissionDistriParIndex[ i-1 ]
                        + NrDistriParams[ EmissionDistriType[ i-1 ] ];

  /* WWTP data */

  WWTPDistriParIndex[0] = 0;

  for ( i=1 ; i < NR_WWTP_PARAMS ; ++i )
    WWTPDistriParIndex[i] =   WWTPDistriParIndex[ i-1 ]
                              + NrDistriParams[ WWTPDistriType[ i-1 ] ];

  TotWWTPDistriPar =   WWTPDistriParIndex[ i-1 ]
                       + NrDistriParams[ WWTPDistriType[ i-1 ] ];

  /* RivClass data */

  RivClassDistriParIndex[0] = 0;

  for ( i=1 ; i < NR_RIVCLASS_PARAMS ; ++i )
    RivClassDistriParIndex[i] =   RivClassDistriParIndex[ i-1 ]
                                  + NrDistriParams[ RivClassDistriType[ i-1 ] ];

  TotRivClassDistriPar =   RivClassDistriParIndex[ i-1 ]
                           + NrDistriParams[ RivClassDistriType[ i-1 ] ];

}



/*
 ********************************************************************
 * P R A C T I C A L    C A L C U L A T I O N S
 ********************************************************************
*/

/*
 * ---------------------------------------------------------------------
 * find the statevar array offset for a specific:
 * statevar, 
 * process, 
 * terminal and
 * submodel
 * ---------------------------------------------------------------------
*/

int oX( int statevar_arg,
        int process_arg,
        int terminal_arg,
        int submodel_arg )
{
  int offset;

  offset = statevar_arg * OneStateVarSize + StateVarsArrayOffset[process_arg];

  switch (ModelClass[process_arg])
  {
    case EMISSION:

      offset += submodel_arg;
      break;

    default:
      offset += terminal_arg * NrSub[process_arg] + submodel_arg;
      break;
  }

  return ( offset );
}

/*
 * ---------------------------------------------------------------------
 * find the summary variables array offset for a specific:
 * statevar, 
 * terminal
 * ---------------------------------------------------------------------
*/

int oS( int statevar_arg,
        int terminal_arg )
{
  return ( statevar_arg * NR_SUMMARY_TERMS + terminal_arg );
}

/*
 * ---------------------------------------------------------------------
 * co-ordinates in A = lower triangular matrix with also zeros on diagonal
 *
 * this function translates 2-D (x,y) co-ordinates into 1-D array coordinates
 *---------------------------------------------------------------------
*/

int Acoordinates( int i_arg,
                  int j_arg )
{
  if ( RowType[i_arg].modelclass == TRANSTREAT && RowType[i_arg].terminal == IN__ )
    return ( AsubMatrixOffset[i_arg] + j_arg);

  else
    return ( AsubMatrixOffset[i_arg] );
}

/*
 * ---------------------------------------------------------------------
 * find a discharge from the data set
 *---------------------------------------------------------------------
*/

int DischargeNumber ( int      from_arg,
                      FloatType segment_ID_arg )
{
  int i;

  /* always an UPWARD search compared to previous */

  for ( i=from_arg ; i<NrSegmentsDischarge ; ++i )
    if ( Distri_DischargeParams[ TotDischargeDistriPar * i ] == segment_ID_arg )
      return i;

  /* didn't find it - retry from beginning */

  for ( i=0 ; i<from_arg ; ++i )
    if ( Distri_DischargeParams[ TotDischargeDistriPar * i ] == segment_ID_arg )
      return i;

  return -1;
}

/*
 * ---------------------------------------------------------------------
 * find a segment from the data set
 * ---------------------------------------------------------------------
*/

int SegmentNumber ( int       from_arg,
                    FloatType segment_ID_arg,
                    int       dir_arg )
{
  int i = 0, j;
  int end1, end2;

  switch ( dir_arg )
  {
    /* UPWARD search compared to previous */

    case _UP_:

      for ( i=from_arg+1 ; i<NrSegments ; ++i )
        if ( Distri_RiverParams[ TotRiverDistriPar * i ] == segment_ID_arg )
          return i;

      /* didn't find it - retry from beginning */

      for ( i=0 ; i<from_arg ; ++i )
        if ( Distri_RiverParams[ TotRiverDistriPar * i ] == segment_ID_arg )
          return i;

      break;

    /* don't know whether it's an UPWARD or DOWNWARD search compared to previous */

    case _DONT_KNOW_:

      for ( j=1, end1=0, end2=0 ; !(end1 && end2) ; ++j )
      {
        if ( !end1 )
          if ( (i = from_arg-j) < 0 )
            end1 = 1;
          else
            if ( Distri_RiverParams[ TotRiverDistriPar * i ] == segment_ID_arg )
              return i;

        if ( !end2 )
          if ( (i = from_arg+j) >= NrSegments )
            end2 = 1;
          else
            if ( Distri_RiverParams[ TotRiverDistriPar * i ] == segment_ID_arg )
              return i;
      }

      break;

    default: break;
  }

  return -1;
}

/*
 * ---------------------------------------------------------------------
 * find an emission data point from the data set
 * ---------------------------------------------------------------------
*/

int EmissionNumber ( FloatType emission_ID_arg )
{
  int i = 0;

  /* downward search from top of list */

  for ( i=0 ; i<NrEmissions ; ++i )
    if ( Distri_EmissionParams [ i * TotEmissionDistriPar ] ==  emission_ID_arg )
      return i;       

  /* didn't find it */
  
  return -1;
}

/*
 * ---------------------------------------------------------------------
 * find a river class data point from the data set
 * ---------------------------------------------------------------------
*/

int RivClassNumber ( FloatType rivclass_ID_arg )
{
  int i = 0;

  /* downward search from top of list */

  for ( i=0 ; i<NrRivClasses ; ++i )
    if ( Distri_RivClassParams [ i * TotRivClassDistriPar ] ==  rivclass_ID_arg )
      return i;       

  /* didn't find it */
  
  return -1;
}

/*
 * ---------------------------------------------------------------------
 * find a WWTP data point from the data set
 * ---------------------------------------------------------------------
*/

int WWTPNumber ( FloatType wwtp_ID_arg )
{
  int i = 0;

  /* downward search from top of list */

  for ( i=0 ; i<NrWWTPs ; ++i )
    if ( Distri_WWTPParams [ i * TotWWTPDistriPar ] ==  wwtp_ID_arg )
      return i;

  /* didn't find it */
  
  return -1;
}


